import io
import math
import uuid
from datetime import date, datetime, timedelta
from pickle import dump, load
from typing import Any, Generator

import pandas as pd
from pydruid.utils.aggregators import longsum
from pydruid.utils.filters import Bound, Dimension, Filter

from common.utils import logging
from common.utils.appinfo.core import AppInfoService
from common.utils.appinfo.data import AppIdProtocol, AppNameStandardPort
from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import Interval, PartialBaseline
from common.utils.checkpoint.config import Config as CheckpointConfig
from common.utils.checkpoint.core import CheckpointService
from common.utils.checkpoint.data import Checkpoint
from common.utils.detector.core import BatchDetector, DetectorType
from common.utils.event.core import EventService
from common.utils.event.data import EventInfo
from common.utils.flow.core import FlowService
from common.utils.flow.data import FlowGroupByQuery
from common.utils.lookup.core import LookupService
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import (
    DetectorRunStatus,
    DetectorStatusLabel,
    ExclusionLabel,
    ExclusionTypeInfo,
    UserConfigInfo,
)
from uncommonport.detector.config import Config
from uncommonport.detector.const import Const
from uncommonport.detector.event import Builder


class UncommonPortBaselineError(Exception):
    pass


class UncommonPortCheckpointError(Exception):
    pass


class UncommonPort(BatchDetector):
    """
    Uncommon port:
    - leverage the flow "L7" appPort mismatch features from the host and report a flow with unusual
      port usage
    - The detector run periodically (with a checkpoint) and pull the flows from druid & select
      the flow with appPort mismatch. Then process the events and push relevant event to psql
    - This detector doesn't use the baseline (since it is a checkpoint based, scanner-like detector)
    """

    logger = logging.get_common_logger(__name__)

    def __init__(
        self,
        baseline_service: BaselineService,
        checkpoint_service: CheckpointService,
        user_config_service: UserConfigService,
        event_service: EventService,
        flow_service: FlowService,
        lookup_service: LookupService,
        app_info_service: AppInfoService,
        now: datetime,
    ):
        super().__init__(baseline_service, checkpoint_service, user_config_service, event_service, now)
        self._flow_service = flow_service
        self._lookup_service = lookup_service
        self._app_info_service = app_info_service
        self._app_info_map = None
        self._now = now
        self._beginning_of_today = datetime.combine(now.date(), datetime.min.time())

    @staticmethod
    def detector_name() -> DetectorType:
        return DetectorType.UNCOMMONLY_USED_PORT

    @staticmethod
    def _baseline_max_retention_time() -> timedelta:
        return timedelta(days=float(Config.BASELINE_MAX_RETENTION_DAYS))

    def _get_day_start_to_now_interval(self) -> Interval:
        # return (start of day today, now)
        return Interval(self._beginning_of_today, self._now)

    @staticmethod
    def get_quadratic_score(value: int, coefficient: float) -> float:
        # (float) getQuadraticScore(2, VM_COUNT_QUADRATIC_COEFFICIENT);
        #   Math.min(Math.sqrt(coefficient*value), MAX_SCORE);
        return min(math.sqrt(coefficient * value), Const.MAX_SCORE)

    def _get_baseline_intervals(self, missing_intervals: list[Interval]) -> list[Interval]:
        """This detector does not use baseline, so returns empty list"""
        return []

    def _generate_baseline(self, site_id: str, interval: Interval) -> PartialBaseline:
        UncommonPort.logger.info(f"ignoring baseline for {interval}")
        raise UncommonPortBaselineError("this detector should not be generating baselines!")

    def _combine_baseline_objects(self, baselines: list[PartialBaseline]) -> io.BytesIO | None:
        UncommonPort.logger.info("ignoring combine baseline")
        return io.BytesIO()

    def _read_checkpoint(self, config: UserConfigInfo) -> io.BytesIO | None:
        checkpoint = self._checkpoint_service.read_checkpoint(self.detector_name().name, Config.ORG_ID, config.site_id)

        if checkpoint is None:
            return None

        buffer = io.BytesIO()
        dump(checkpoint, buffer)
        buffer.seek(0)

        return buffer

    def _create_checkpoint(self, config: UserConfigInfo) -> io.BytesIO | None:
        """
        create a default checkpoint with datetime = now-1day, empty df
        :param config:
        :return:
        """
        today_midnight = datetime.combine(date.today(), datetime.min.time())

        # create empty df (serialize as parquet)
        buffer = io.BytesIO()
        empty_df = pd.DataFrame()
        empty_df.to_parquet(buffer)
        buffer.seek(0)

        default_checkpoint = Checkpoint(today_midnight, "1.0", buffer)
        checkpoint_name = CheckpointService.get_prefix(self.detector_name().name, Config.ORG_ID, config.site_id)

        # create checkpoint file in minio
        self._checkpoint_service.get_or_create_bucket()

        # minio client uses read() to get the bytes to write, so need to set cursor to beginning of the buffer
        default_checkpoint.data.seek(0)

        # turn checkpoint into a bytesIO
        checkpoint_buffer = io.BytesIO()
        dump(default_checkpoint, checkpoint_buffer)
        checkpoint_buffer.seek(0)

        success = self._checkpoint_service.put_object(
            CheckpointConfig.BASELINE_BUCKET,
            self.detector_name().name,
            Config.ORG_ID,
            config.site_id,
            checkpoint_buffer,
        )

        UncommonPort.logger.info(
            "created checkpoint in bucket {} with name: {}({})".format(
                CheckpointConfig.BASELINE_BUCKET, checkpoint_name, success
            )
        )

        return self._read_checkpoint(config)

    def _get_check_firewall_rule_from_config(self, config: UserConfigInfo) -> bool:
        firewall_rule_setting = config.settings.get_firewall_rule_setting()
        if firewall_rule_setting is None:
            UncommonPort.logger.warning("got None when trying to read firewall rule setting, defaulting to False")
            return False
        return firewall_rule_setting.check_firewall_rule

    def _get_excluded_computes_from_config(self, config: UserConfigInfo) -> list[str]:
        if config.exclusion_settings == {}:
            return []

        excluded_computes = []
        for setting in config.exclusion_settings.exclusions:
            # technically, it should be ExclusionLabel.DESTINATION or SOURCE but the existing API
            # is using NONE. In order to avoid changing the API contract, we allow both
            # DESTINATION, SOURCCE or NONE (existing behavior)
            if (
                setting.label == ExclusionLabel.DESTINATION
                or setting.label == ExclusionLabel.SOURCE
                or setting.label == ExclusionLabel.NONE
            ):
                # setting.excluded_items_info.value can technically be any type, but here we're assuming it's a list of str
                # type: ignore (suppress linter warning)
                excluded_entity_ids: list[str] = setting.excluded_items_info.value
                if setting.type == ExclusionTypeInfo.VM:
                    excluded_computes += excluded_entity_ids
                elif setting.type == ExclusionTypeInfo.GROUP:
                    excluded_computes += self._lookup_service.groups_to_computes_lookup(
                        config.site_id, excluded_entity_ids
                    )
        return excluded_computes

    def _get_flow_with_port_mismatch(
        self, interval: Interval, checkpoint_interval: Interval, site_id: str
    ) -> pd.DataFrame:
        return self._execute_flow_query(interval, checkpoint_interval, site_id)

    def _execute_flow_query(self, interval: Interval, checkpoint_interval: Interval, site_id: str) -> pd.DataFrame:
        filter_src_compute_type = Const.SRC_COMPUTE_TYPE
        filter_dst_compute_type = Const.DST_COMPUTE_TYPE
        filter_src_site = Const.SRC_SITE_ID
        filter_dst_site = Const.DST_SITE_ID

        # or(and(src_compute, src_site, src_mismatch > 0), and(dst_compute, dst_site, dst_mismatch > 0))
        filter_src_compute_site = [
            Dimension(filter_src_compute_type) == Const.VM_COMPUTE_TYPE,
            Dimension(filter_src_site) == site_id,
            Bound(Const.SRC_APP_PORT_MISMATCHED, "0", lowerStrict=True),
            Dimension("isSrcInfoPresent") == True,  # noqa
        ]
        filters_src = Filter(type="and", fields=filter_src_compute_site)

        filter_dst_compute_site = [
            Dimension(filter_dst_compute_type) == Const.VM_COMPUTE_TYPE,
            Dimension(filter_dst_site) == site_id,
            Bound(Const.DST_APP_PORT_MISMATCHED, "0", lowerStrict=True),
            Dimension("isDstInfoPresent") == True,  # noqa
        ]
        filters_dst = Filter(type="and", fields=filter_dst_compute_site)

        or_filters = Filter(type="or", fields=[filters_src, filters_dst])

        # add the bounded filter: "checkpoint start < flow lastStartTime < checkpoint end"
        checkpoint_filter = Bound(
            Const.LAST_START_TIME,
            lower=round(checkpoint_interval.start_time.timestamp()),
            upper=round(checkpoint_interval.end_time.timestamp()),
        )

        # final "and" filter
        filters = [or_filters, checkpoint_filter]

        # add filter to select flows that happen *after* the checkpoint date
        #
        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_ALL_GRANULARITY,
            intervals=[interval],
            dimensions=[
                Const.SRC_VM_ID,
                Const.DST_VM_ID,
                Const.DST_PORT,
                Const.PROTOCOL,
                Const.APP_ID,
                Const.DST_INSECURE_APP_VERSION,
                Const.IS_SRC_L7_RULE_HIT,
                Const.IS_DST_L7_RULE_HIT,
            ],
            aggregations={
                Const.SRC_APP_PORT_MISMATCHED: longsum(Const.SRC_APP_PORT_MISMATCHED),
                Const.DST_APP_PORT_MISMATCHED: longsum(Const.DST_APP_PORT_MISMATCHED),
            },
            post_aggregations={},
            filter=Filter(type="and", fields=filters),
            context={"queryId": f"NTA:{self.detector_name()}:flow-app:{uuid.uuid4()}"},
        )

        UncommonPort.logger.info(f"executing flow query:\n{query.pprint()}")
        result = self._flow_service.group_by(query)

        UncommonPort.debug_log_df(result, site_id, "flow-app")

        return result

    def _get_appinfo_map(self) -> dict[AppIdProtocol, AppNameStandardPort]:
        if self._app_info_map is None:
            self._app_info_map = self._app_info_service.get_app_info()

        return self._app_info_map

    def _create_event(self, result_row: pd.Series, site_id: str, compute_to_ip_lookup: dict[str, str]) -> EventInfo:
        # Mark detection event time (epoch mill)
        detection_time = int(self._now.timestamp() * 1000)

        # app_info map
        std_app_info_map = self._get_appinfo_map()
        std_app_info_key = AppIdProtocol(int(result_row[Const.APP_ID]), result_row[Const.PROTOCOL])

        std_app_display_name = (
            std_app_info_map[std_app_info_key].display_name
            if std_app_info_key in std_app_info_map
            else Const.UNKNOWN_APP_DISPLAY_NAME
        )

        std_app_expected_ports = (
            std_app_info_map[std_app_info_key].expected_ports
            if std_app_info_key in std_app_info_map
            else Const.UNKNOWN_APP_EXPECTED_PORTS
        )

        event = (
            Builder()
            .detector_type(self.detector_name())
            # use detector name as default description
            .description(self.detector_name().name)
            .detection_time(detection_time)
            .silent(result_row[Const.IS_SILENT])
            .org_id(Config.ORG_ID)
            .site_id(site_id)
            .compute_ip_map(compute_to_ip_lookup)
            .app_id(int(result_row[Const.APP_ID]))
            .protocol(result_row[Const.PROTOCOL])
            .dst_port(result_row[Const.DST_PORT])
            .app_name(std_app_display_name)
            .discovered_port(int(result_row[Const.DST_PORT]))
            .expected_ports(std_app_expected_ports)
            .vulnerable_app_version(result_row[Const.DST_INSECURE_APP_VERSION])
            .src_computes(result_row[Const.SRC_VM_ID])
            .dst_computes(result_row[Const.DST_VM_ID])
            .viz_filter([result_row[Const.APP_ID]], [result_row[Const.DST_PORT]])
            .confidence(self.get_quadratic_score(Const.DEFAULT_SCORE_VALUE, Const.VM_COUNT_QUADRATIC_COEFFICIENT))
            .build()
        )

        return event

    def _filter_result(self, user_config: UserConfigInfo, result: pd.DataFrame) -> pd.DataFrame:
        filtered_result = result  # default

        check_firewall_rule = self._get_check_firewall_rule_from_config(user_config)

        # If a L7 firewall rule is hit and user settings suggest to check firewall rules,
        # then we shouldn't create an event.
        # keep the event if the above condition is not met
        # TODO: druid return true/false as string (instead of boolean). So converting column via astype(bool)
        #       will not do the proper filtering
        UncommonPort.logger.info(f"filtering result based on checkFirewallRule option {check_firewall_rule}")
        if check_firewall_rule:
            filtered_result = result[
                (result[Const.IS_DST_L7_RULE_HIT] == "false") & (result[Const.IS_SRC_L7_RULE_HIT] == "false")
            ]

        UncommonPort.debug_log_df(filtered_result, user_config.site_id, "check-fw-filter")

        # Find excluded vm and filter out the results
        excluded_computes = self._get_excluded_computes_from_config(user_config)
        UncommonPort.logger.info(f"filtering result based exclusion {excluded_computes}")

        filtered_result = filtered_result.loc[
            ~(result[Const.SRC_VM_ID].isin(excluded_computes) | result[Const.DST_VM_ID].isin(excluded_computes))
        ]

        UncommonPort.debug_log_df(filtered_result, user_config.site_id, "exclude-compute")

        return filtered_result

    def _detect_events(
        self, user_config: UserConfigInfo, baseline: io.BytesIO, checkpoint: io.BytesIO | None
    ) -> Generator[EventInfo, Any, Any]:
        # 1. read the checkpoint, specify flow query interval
        if checkpoint is None:
            raise UncommonPortCheckpointError(
                "got None for checkpoint during event detection, expected checkpoint object"
            )
        c = load(checkpoint)

        # 2a. select cfv within interval & identify flow with port-mismatched
        now = self._now

        # cap the limit of query (max(last_checkpoint, or now-120min)
        checkpoint_start_limit = max(
            c.time, now.replace(microsecond=0) - timedelta(minutes=float(Config.MAX_CHECKPOINT_LIMIT_IN_MIN))
        )
        UncommonPort.logger.info(f"checkpoint interval: {checkpoint_start_limit} to {now}")

        # give a small extra window for late flow:
        # by reducing the start_time by expected flow delays (average 7.5min)
        #
        # The reason is that flow could come in around 5+mins late (due to lags
        # from rest of the system). The detection window might miss these late
        # flows that falls outside (esp. right before) the window
        #
        # The detection window now overlaps by ~7.5min
        # The windows overlaps at (*)
        #  <---w1--->
        #          <---w2--->
        #                   <---w2--->
        # -*0------*1-------*2-------3
        #
        # The trade off is that there is a chance of duplicate events
        # (which nta-server will deduplicate)
        #
        # Example to demostrate the issue:
        # 0.0----1.0---(flow time=0.9, insert at 1.1)-----2.0 (timeline)
        #
        # The current detection window is [1.0->2.0].
        # If flow f shows up late (e.g. lastStartTime=0.9 but insert at 1.1),
        # the detector window [1.0->2.0] would ignore flow f (since lastStartTime is outside of [1.0->2.0])
        #
        # But the previous window [0.0-1.0] will not see f neither since f get inserted late at 1.1,
        # which is non-existence during time [0.0-1.0]
        #
        checkpoint_start_limit = checkpoint_start_limit - timedelta(seconds=float(Config.MAX_LATE_FLOW_LIMIT_IN_SEC))

        # druid "interval" is based on __time column and not the lastStartTime (the actual time of the flow)
        # checkpoint_interval represent the actual time we use in the filter parameter; Unlike __time which
        # get rounds to the nearest hour
        checkpoint_interval = Interval(checkpoint_start_limit, now)
        UncommonPort.logger.info(
            f"adjusted checkpoint interval: {checkpoint_interval}. " f"tolerance: {Config.MAX_LATE_FLOW_LIMIT_IN_SEC}"
        )

        # interval represent the druid's query __time column which get round down
        # but since druid query requires interval in its query, we create a large enough
        # bound by rounding down/up of the checkpoint_interval start/end
        interval = Interval(checkpoint_interval.start_time, checkpoint_interval.end_time)

        # round down start, round up end time
        interval.start_time = interval.start_time.replace(minute=0, second=0, microsecond=0)
        interval.end_time = interval.end_time.replace(minute=0, second=0, microsecond=0) + timedelta(hours=1)
        UncommonPort.logger.info(f"adjusted query interval: {interval}")

        """
        | dstVmId | isSrcL7RuleHit   | protocol   |   dstInsecureAppVersion |   dstPort | srcVmId  |   appId |
        |:--------|:-----------------|:-----------|------------------------:|----------:|:---------|--------:|
        | vm-0    | false            | TCP        |                       0 |        64 | vm-2     |      48 |
        | vm-1    | false            | TCP        |                       0 |        44 | vm-3     |      48 |

        |   srcAppPortMismatched |   dstAppPortMismatched | isDstL7RuleHit   | timestamp                |
        |-----------------------:|-----------------------:|:-----------------|:-------------------------|
        |                      1 |                      1 | false            | 2023-05-25T16:20:48.883Z |
        |                      1 |                      1 | false            | 2023-05-25T16:20:48.883Z |
        """
        # 2b. Run the main port mismatch query
        result = self._get_flow_with_port_mismatch(interval, checkpoint_interval, user_config.site_id)
        if len(result.index) <= 0:
            UncommonPort.logger.info(f"port mismatch result is empty {interval} for {user_config.site_id}")
            return

        # 3. Filter out results (settings, exclusion)
        result = self._filter_result(user_config, result)
        if len(result.index) <= 0:
            UncommonPort.logger.info(f"port mismatch result is empty {interval} for {user_config.site_id}")
            return

        # 4a. prepare the appinfo lookup
        std_app_info_map = self._get_appinfo_map()
        if not std_app_info_map:
            UncommonPort.logger.warn("Empty app info map. Skipping detection run. Please check the psql tables")
            return

        # 4b. compare result vs standard port list (std_app_info_map)
        #     only keep result row where the AppId is in the std app info map
        std_app_id_list = list({appId_protocol.app_id for appId_protocol in std_app_info_map.keys()})
        std_app_id_list.sort()
        result = result[result[Const.APP_ID].astype(int).isin(std_app_id_list)]

        # skip flows with unknown app id
        if len(result) <= 0:
            UncommonPort.logger.error(
                "result contains unknown / invalid app id. result: {result[Const.APP_ID]}"
                "known app id: {std_app_id_list}. No event will be generated."
                "Consider checking the psql table for make sure the app id are valid"
            )
            return

        # 4b: add column to indicate if result['dstPort'] is not in std_info_map port list
        def is_expected_port(res, std_app: dict[AppIdProtocol, AppNameStandardPort]) -> bool:
            key = AppIdProtocol(int(res[Const.APP_ID]), res[Const.PROTOCOL])

            if key not in std_app:
                UncommonPort.logger.error("missing appid, protocol key {key} in std app map")
                return False

            return int(res[Const.DST_PORT]) in std_app[key].expected_ports

        result[Const.IS_PORT_EXPECTED] = result.apply(is_expected_port, args=(std_app_info_map,), axis=1)

        # 4c: filter row if port is unexpected
        result = result.loc[~result[Const.IS_PORT_EXPECTED]]

        # 4d: group by the dedup key to prevent publishing events that will anyway get deduped
        result = result.groupby(["appId", "dstPort"], as_index=False).agg(
            {
                Const.SRC_VM_ID: Const.AGG_LIST_TRANSFORM_WITH_LIMIT_WITHOUT_DUP,
                Const.DST_VM_ID: Const.AGG_LIST_TRANSFORM_WITH_LIMIT_WITHOUT_DUP,
                Const.DST_PORT: Const.AGG_FIRST,
                Const.PROTOCOL: Const.AGG_FIRST,
                Const.DST_INSECURE_APP_VERSION: Const.AGG_FIRST,
                Const.APP_ID: Const.AGG_FIRST,
            }
        )

        # 5: uncommonport is not silent, since it doesn't have sensitivity
        result[Const.IS_SILENT] = False

        UncommonPort.debug_log_df(result, user_config.site_id, "unexpected-ports")

        if len(result.index) == 0:
            UncommonPort.logger.warn("empty result. No event will be generated")
            return

        UncommonPort.debug_log_df(result, user_config.site_id, "Uncommon Event")

        # prepare a compute->ip lookup table for all events
        compute_to_ip_lookup = self._lookup_service.computes_to_ip_lookup(user_config.site_id, interval)
        UncommonPort.logger.info(f"preparing compute to ip lookup (size:{len(compute_to_ip_lookup)})")

        # 6. generate events
        for _, row in result.iterrows():
            UncommonPort.logger.info(f"creating {self.detector_name()} event")
            yield self._create_event(row, user_config.site_id, compute_to_ip_lookup)

    def _update_checkpoint(self, user_config: UserConfigInfo) -> io.BytesIO | None:
        new_checkpoint = self._checkpoint_service.read_checkpoint(
            self.detector_name().name, Config.ORG_ID, user_config.site_id
        )
        if new_checkpoint is None:
            UncommonPort.logger.warning("unexpectedly got None checkpoint when trying to update")
            return None

        # update the timestamp for the checkpoint with the (true) current time
        # round to nearest second
        new_checkpoint.time = datetime.now().replace(microsecond=0)

        # 8. convert checkpoint to io buffer
        checkpoint_buffer = io.BytesIO()
        dump(new_checkpoint, checkpoint_buffer)
        checkpoint_buffer.seek(0)

        return checkpoint_buffer

    @staticmethod
    def debug_log_df(df: pd.DataFrame, site_id: str, message: str = ""):
        with pd.option_context("display.max_rows", None, "display.max_columns", None, "display.precision", 3):
            UncommonPort.logger.debug(
                "================================================================================"
            )
            UncommonPort.logger.debug(f"{message}-{site_id} result:")
            UncommonPort.logger.debug(df.to_markdown()) if not df.empty else UncommonPort.logger.debug("<EMPTY DF>")
            UncommonPort.logger.debug(
                "================================================================================"
            )

    def _get_detector_run_status(
        self, status_label: DetectorStatusLabel, site_id: str, last_successful_run_time: int, message: str = ""
    ) -> DetectorRunStatus:
        UncommonPort.logger.debug(f"Setting detector run status for {self.detector_name().name}")
        last_successful_run_time = last_successful_run_time
        checkpoint = self._checkpoint_service.read_checkpoint(self.detector_name().name, Config.ORG_ID, site_id)
        checkpoint_time = 0
        if checkpoint is not None:
            checkpoint_time = int(checkpoint.time.timestamp() * 1000)
        return DetectorRunStatus(
            status_label=status_label,
            last_successful_run_time=last_successful_run_time,
            checkpoint_time=checkpoint_time,
            message=message,
        )
