import json
from typing import Self

import numpy as np

import openapi.models.base_anomaly_event_detail  # needed to avoid circular import for any subclasses  # noqa: F401
from common.utils import logging
from common.utils.detector.core import DetectorType
from common.utils.event.data import (
    EventInfo,
    EventKey,
    EventMetadata,
    JsonValue,
    ListString,
    MergePlan,
    MergePolicy,
    Score,
    Tactic,
    Technique,
)
from common.utils.vizfilter.core import (
    CompositeVizFilter,
    Condition,
    Criteria,
    VizFilter,
    VizFilterOption,
)
from openapi.models.non_standard_port_usage_event_detail import (
    NonStandardPortUsageEventDetail,
)
from uncommonport.detector.const import Const
from vmware.nsx.intelligence.nta.v1 import nta_pb2


class Builder:
    logger = logging.get_common_logger(__name__)
    """
    convert varies metrics about uncommon port event into the API object
    """
    _detector_type: DetectorType
    _description: str
    _detection_time: int
    _src_computes: list[str]
    _dst_computes: list[str]
    _hotspots: list[str]
    _org_id: str
    _site_id: str
    _silent: bool
    _confidence: int
    _final_viz_filter: CompositeVizFilter
    _app_id: int
    _app_name: str
    _dst_port: int
    _discovered_port: int
    _expected_ports: list[int]
    _protocol: str
    _vulnerable_app_version: str
    _compute_to_ip: dict[str, str]
    _is_src_l7_rule_hit: bool
    _is_src_l7_rule_hit: bool

    def detector_type(self, detector_type: DetectorType) -> Self:
        self._detector_type = detector_type
        return self

    def description(self, description: str) -> Self:
        self._description = description
        return self

    def org_id(self, org_id: str) -> Self:
        self._org_id = org_id
        return self

    def site_id(self, site_id: str) -> Self:
        self._site_id = site_id
        return self

    def app_id(self, app_id: int) -> Self:
        self._app_id = app_id
        return self

    def dst_port(self, dst_port: int) -> Self:
        self._dst_port = dst_port
        return self

    def protocol(self, protocol: str) -> Self:
        self._protocol = protocol
        return self

    def app_name(self, app_name: str) -> Self:
        self._app_name = app_name
        return self

    def vulnerable_app_version(self, vulnerable_app_version) -> Self:
        self._vulnerable_app_version = vulnerable_app_version
        return self

    def confidence(self, score) -> Self:
        self._confidence = np.floor(score).astype(int)  # score need to be int
        return self

    def compute_ip_map(self, compute_to_ip: dict[str, str]) -> Self:
        self._compute_to_ip = compute_to_ip
        return self

    def discovered_port(self, discovered_port: int) -> Self:
        self._discovered_port = discovered_port
        return self

    def expected_ports(self, expected_ports: list[int]) -> Self:
        self._expected_ports = expected_ports
        return self

    def src_computes(self, compute_ids: list[str]) -> Self:
        self._src_computes = compute_ids
        return self

    def dst_computes(self, compute_ids: list[str]) -> Self:
        self._dst_computes = compute_ids
        return self

    def detection_time(self, detection_time: int) -> Self:
        self._detection_time = detection_time
        return self

    def silent(self, is_silent: bool) -> Self:
        self._silent = is_silent
        return self

    def viz_filter(self, app_ids: list[int], ports: list[int]) -> Self:
        # flow dstport
        flow_dstport_filter = VizFilter(
            key=VizFilterOption.FLOW_DSTPORT,
            values=[str(port) for port in ports],
            criteria=Criteria.EQ,
            condition=Condition.OR,
        )

        app_id_filter = VizFilter(
            key=VizFilterOption.FLOW_APPID,
            values=[str(app_id) for app_id in app_ids],
            criteria=Criteria.EQ,
            condition=Condition.OR,
        )

        self._final_viz_filter = CompositeVizFilter([flow_dstport_filter, app_id_filter], Condition.AND)

        return self

    def build_details(self) -> NonStandardPortUsageEventDetail:
        return NonStandardPortUsageEventDetail(
            report_type=self._detector_type.name,  # report type has same name as detector
            expected_port=self._expected_ports,
            discovered_port=self._discovered_port,
            app_id=self._app_id,
            app_name=self._app_name,
            protocol=self._protocol,
            vulnerable_app_version=self._vulnerable_app_version,
        )

    def build(self) -> EventInfo:
        # NOTE: take the first IP from the computes list since the agg function used below in the merge plan is KEEP_ORIGINAL
        src_ip = ""
        dst_ip = ""
        if len(self._src_computes) > 0:
            src_ip = self._compute_to_ip[self._src_computes[0]] if self._src_computes[0] in self._compute_to_ip else ""
        if len(self._dst_computes) > 0:
            dst_ip = self._compute_to_ip[self._dst_computes[0]] if self._dst_computes[0] in self._compute_to_ip else ""

        # metadata
        event_metadata = EventMetadata(
            kv_store={
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_COMPUTES): ListString(self._src_computes),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_COMPUTES): ListString(self._dst_computes),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_HOTSPOTS): ListString(
                    [*self._src_computes, *self._dst_computes]
                ),
                EventKey.EVENT_METADATA_REPORT_KEY: JsonValue(json.dumps(self.build_details().to_dict())),
                EventKey.EVENT_METADATA_DESCRIPTION_KEY: JsonValue(self._description),
                EventKey.EVENT_METADATA_VIZ_FILTER_KEY: JsonValue(self._final_viz_filter.to_json_str()),
                # NDR needs 1 src and 1 dst ip associated with each event,
                # look up ip based on src/dst vm (replace fetchSrcDstIps)
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_IPS): ListString([src_ip]),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_IPS): ListString([dst_ip]),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_PROTOCOL): ListString([self._protocol]),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_PORT): ListString([str(self._dst_port)]),
            }
        )

        # merge_plan
        merge_plan = MergePlan(
            # porting note: the old code is iterating through the dst/src compute as key, but we now use the
            #               compute_id for each builder invocation
            # TODO: we might need to dedup based on dst_compute also (for NDR)
            #       e.g. add (self._dst_compute) and remove the merge op for dst compute
            # NOTE: we group by the dedup key before publishing the event, so if you're changing the dedup key
            #       here, please also check out core.py _detect_events() step 4d and update the group by there as well
            dedup_key="{}_{}_{}_{}".format(
                Const.NSX_I, str(DetectorType.UNCOMMONLY_USED_PORT), str(self._app_id), str(self._dst_port)
            ),
            metadata_key_merge_operation={
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_COMPUTES): MergePolicy.SET_UNION_WITH_MAX,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_COMPUTES): MergePolicy.SET_UNION_WITH_MAX,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_HOTSPOTS): MergePolicy.SET_UNION_WITH_MAX,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_IPS): MergePolicy.KEEP_ORIGINAL,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_IPS): MergePolicy.KEEP_ORIGINAL,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_PROTOCOL): MergePolicy.KEEP_ORIGINAL,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_PORT): MergePolicy.KEEP_ORIGINAL,
            },
        )

        event_info = EventInfo(
            version="v1.1",
            detector_name=self._detector_type,
            detection_start_time_epoch_milli=self._detection_time,
            detection_end_time_epoch_milli=self._detection_time,
            score=Score(self._confidence, severity=20),
            event_metadata=event_metadata,
            merge_plan=merge_plan,
            mitre_tactic=Tactic.COMMAND_AND_CONTROL,
            mitre_technique=Technique.NON_STANDARD_PORT,
            silent=self._silent,
            site_ids=[self._site_id],
            org_id=self._org_id,
            # affected compute is not expected for this event, since there's generally no compute affected
            affected_compute=None,
            is_attacker=None,
            is_client=None,
            data_flows=None,
        )

        return event_info
