import io
import unittest
from datetime import date, datetime, timedelta
from pickle import load
from unittest.mock import Mock

import pandas as pd

from common.utils.appinfo.core import AppInfoService
from common.utils.appinfo.data import AppIdProtocol, AppNameStandardPort
from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import Interval, PartialBaseline
from common.utils.checkpoint.core import CheckpointService
from common.utils.checkpoint.data import Checkpoint
from common.utils.detector.core import DetectorType
from common.utils.event.core import EventService
from common.utils.event.data import Score
from common.utils.flow.core import FlowService
from common.utils.lookup.core import LookupService
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import (
    AnomalyExclusionSettings,
    DetectorRunStatus,
    DetectorSettings,
    DetectorStatusLabel,
    UserConfigInfo,
)
from uncommonport.detector.config import Config
from uncommonport.detector.const import Const
from uncommonport.detector.core import UncommonPort
from uncommonport.detector.event import Builder


class TestCore(unittest.TestCase):
    def setUp(self):
        self._mock_flow_service = Mock(FlowService)
        self._mock_checkpoint_service = Mock(CheckpointService)
        self._mock_app_info_service = Mock(AppInfoService)
        self.detector = UncommonPort(
            Mock(BaselineService),
            self._mock_checkpoint_service,
            Mock(UserConfigService),
            Mock(EventService),
            self._mock_flow_service,
            Mock(LookupService),
            self._mock_app_info_service,
            datetime.now(),
        )

    def test_detector_name(self):
        self.assertEqual(DetectorType.UNCOMMONLY_USED_PORT, self.detector.detector_name())

    def test_combine_baseline(self):
        baselines = []
        for _ in range(10):
            baselines.append([[f"VM {i}", i * 10 % 100] for i in range(10)])
        baselines = [pd.DataFrame(baseline, columns=["dstVmId", "cardinalityCount"]) for baseline in baselines]
        for baseline in baselines:
            baseline.set_index("dstVmId", inplace=True)

        partial_baselines = []
        for baseline in baselines:
            buffer = io.BytesIO()
            baseline.to_parquet(buffer)
            partial_baselines.append(PartialBaseline(Interval(datetime.now(), datetime.now()), buffer))

        self.detector._combine_baseline_objects(partial_baselines)

    def test_build_event(self):
        event = (
            Builder()
            .src_computes(["vm1"])
            .compute_ip_map({})
            .dst_computes(["vm2"])
            .detector_type(DetectorType.UNCOMMONLY_USED_PORT)
            .expected_ports([80, 443])
            .discovered_port(22)
            .app_id(1)
            .app_name("http")
            .protocol("TCP")
            .vulnerable_app_version("v1")
            .description("test event")
            .viz_filter([1], [22])
            .dst_port(22)
            .detection_time(int(datetime.now().timestamp()))
            .confidence(50)
            .silent(False)
            .site_id("site 1")
            .org_id("org 1")
            .build()
        )

        self.assertEqual(["site 1"], event.site_ids)
        self.assertEqual(Score(50, 20), event.score)

    def test_execute_flow_query(self):
        self._mock_flow_service.group_by = lambda _: pd.DataFrame()

        today = datetime.combine(date.today(), datetime.min.time())
        result = self.detector._execute_flow_query(
            Interval(today - timedelta(days=1), today), Interval(today - timedelta(days=1), today), "site 1"
        )

        self.assertEqual(0, len(result.index))

    def test_create_checkpoint(self):
        user_config_info = UserConfigInfo(
            "site 1",
            DetectorType.UNCOMMONLY_USED_PORT.name,
            "org 1",
            "v1.1",
            0,
            True,
            False,
            "test",
            False,
            50,
            [],
            [],
            DetectorSettings([]),
            AnomalyExclusionSettings([]),
            DetectorRunStatus(DetectorStatusLabel.SUCCESS),
        )

        self._mock_checkpoint_service.get_prefix = lambda _x, _y, _z: "prefix/"
        self._mock_checkpoint_service.read_checkpoint = lambda _x, _y, _z: Checkpoint(
            datetime.now(), "v1.1", io.BytesIO()
        )

        result = self.detector._create_checkpoint(user_config_info)

        self.assertNotEqual(None, result)
        if result is not None:
            checkpoint_result = load(result)
            self.assertEqual("v1.1", checkpoint_result.version)

    def test_detect_events(self):
        user_config_info = UserConfigInfo(
            "site 1",
            DetectorType.UNCOMMONLY_USED_PORT.name,
            "org 1",
            "v1.1",
            0,
            True,
            False,
            "test",
            False,
            50,
            [],
            [],
            DetectorSettings([]),
            AnomalyExclusionSettings([]),
            DetectorRunStatus(DetectorStatusLabel.SUCCESS),
        )

        self._mock_checkpoint_service.get_prefix = lambda _x, _y, _z: "prefix/"
        self._mock_checkpoint_service.read_checkpoint = lambda _x, _y, _z: Checkpoint(
            datetime.now(), "v1.1", io.BytesIO()
        )

        flow_query_data = [
            ["vm1", "vm2", 22, "TCP", "1", "1", True, False, 1, 2],
            ["vm1", "vm2", 22, "TCP", "1", "1", True, False, 1, 2],
            ["vm1", "vm2", 22, "TCP", "1", "1", True, False, 1, 2],
            ["vm1", "vm2", 22, "TCP", "1", "1", True, False, 1, 2],
        ]

        flow_query_df = pd.DataFrame(
            flow_query_data,
            columns=[
                "srcVmId",
                "dstVmId",
                "dstPort",
                "protocol",
                "appId",
                "dstInsecureAppVersion",
                "isSrcL7RuleHit",
                "isDstL7RuleHit",
                "srcAppPortMismatched",
                "dstAppPortMismatched",
            ],
        )

        self._mock_flow_service.group_by = lambda _: flow_query_df

        app_info_map = {AppIdProtocol(1, "TCP"): AppNameStandardPort("SSH", [22])}

        self._mock_app_info_service.get_app_info = lambda: app_info_map

        results = [
            _
            for _ in self.detector._detect_events(
                user_config_info, io.BytesIO(), self.detector._create_checkpoint(user_config_info)
            )
        ]

        self.assertEqual(0, len(results))

        checkpoint = self.detector._update_checkpoint(user_config_info)
        if checkpoint is not None:
            self.assertEqual("v1.1", load(checkpoint).version)
        else:
            self.fail("checkpoint shouldn't be None")

    def test_event_aggregation(self):
        # test dedup property
        test_series = pd.Series(["a", "b", "b", "c", "c", "c", "c", "a"])
        result = Const.AGG_LIST_TRANSFORM_WITH_LIMIT_WITHOUT_DUP(test_series)
        result.sort()
        self.assertEquals(["a", "b", "c"], result)

        # test limit property
        Config.AGG_SET_TRANSFORM_LIMIT = 5
        test_series = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9])
        result = Const.AGG_LIST_TRANSFORM_WITH_LIMIT_WITHOUT_DUP(test_series)
        self.assertEquals(Config.AGG_SET_TRANSFORM_LIMIT, len(result))


def test_get_detector_status(self):
    self.mock_checkpoint_service = Mock(CheckpointService())
    self.mock_checkpoint_service.read_checkpoint = lambda _x, _y, _z: Checkpoint(
        datetime(year=2023, month=1, day=1), "1", io.BytesIO()
    )
    original_checkpoint_service = self.detector._checkpoint_service
    self.detector._checkpoint_service = self.mock_checkpoint_service

    # FAILURE case
    detector_run_status = self.detector._get_detetor_run_status(DetectorStatusLabel.FAILURE, "123", 1234, "failed")
    self.assertEqual(DetectorStatusLabel.FAILURE, detector_run_status.status_label)

    # SUCCESS case
    detector_run_status = self.detector._get_detetor_run_status(DetectorStatusLabel.SUCCESS, "123", 1234, "success")
    self.assertEqual(DetectorStatusLabel.SUCCESS, detector_run_status.status_label)
    self.detector._checkpoint_service = original_checkpoint_service
